@@ -27,6 +27,7 @@ import struct
 import threading
 from enum import Enum
 from functools import wraps
+from collections import defaultdict
 
 from fdb.impl import KeySelector
 
@@ -37,15 +38,19 @@ from found._fdb import ffi
 log = logging.getLogger(__name__)
 
 
-CLIENT_VERSION = 600
-
-
 class BaseFound:
     """Base class for all found classes"""
 
     pass
 
 
+class Hook(set):
+
+    async def run(self, tx):
+        for func in self:
+            func(tx)
+
+
 class FoundException(Exception):
     """Base class for all found exceptions"""
 
@@ -67,7 +72,7 @@ class FoundError(FoundException):
 
 
 def strinc(key):
-    """Compute the smallest key that does not start with key"""
+    """Compute the smallest bytes sequence that does not start with key"""
     key = key.rstrip(b"\xff")
     if len(key) == 0:
         raise ValueError("Key must contain at least one byte not equal to 0xFF.")
@@ -136,7 +141,7 @@ class StreamingMode(Enum):
 def on_transaction_get_read_version(fdb_future, aio_future):
     aio_future = ffi.from_handle(aio_future)
     pointer = ffi.new("int64_t *")
-    error = lib.fdb_future_get_version(fdb_future, pointer)
+    error = lib.fdb_future_get_int64(fdb_future, pointer)
     if error == 0:
         _loop.call_soon_threadsafe(aio_future.set_result, pointer[0])
     else:
@@ -207,8 +212,9 @@ def on_transaction_get_range(fdb_future, aio_future):
         # total count of buffers for this key-value array
         # total = count[0] * 2
         # free = on_transaction_get_range_free(fdb_future, total)
+        copy = kvs[0][0:count[0]]
 
-        for kv in kvs[0][0:count[0]]:
+        for kv in copy:
             # XXX: manual unpacking because cffi doesn't known about packing
             # https://bitbucket.org/cffi/cffi/issues/364/make-packing-configureable
             memory = ffi.buffer(ffi.addressof(kv), 24)
@@ -245,7 +251,7 @@ def get_value(obj):
         value = obj.as_foundationdb_value()
     except AttributeError:
         value = obj
-    assert isinstance(obj, bytes)
+    assert isinstance(value, bytes)
     return value
 
 
@@ -374,6 +380,7 @@ class Transaction(BaseTransaction):
     def __init__(self, pointer, database):
         super().__init__(pointer, database, False)
         self.snapshot = BaseTransaction(pointer, database, True)
+        self.state = defaultdict(dict)
 
     def __del__(self):
         # because self.pointer is shared with self.snapshot, we
@@ -400,7 +407,7 @@ class Transaction(BaseTransaction):
             end = get_key(end)
         lib.fdb_transaction_clear_range(self._pointer, begin, len(begin), end, len(end))
 
-    async def commit(self):
+    async def _commit(self):
         fdb_future = lib.fdb_transaction_commit(self._pointer)
         aio_future = _loop.create_future()
         handle = ffi.new_handle(aio_future)
@@ -474,7 +481,7 @@ def transactional(func):
         # XXX: 'tr' can not be passed as a keyword
         index = spec.args.index("tr")
     except ValueError:
-        msg = "the decorator @transactional expect one of the argument to be name 'tr'"
+        msg = "Expected `tr` as one positional argument."
         raise NameError(msg)
 
     @wraps(func)
@@ -485,15 +492,18 @@ def transactional(func):
             return out
         else:
             tx = db_or_tx._create_transaction()
+            await db_or_tx.on_transaction_begin.run(tx)
             # replace db with tx *ahem* tr
             args = list(args)
             args[index] = tx
             while True:
                 try:
                     out = await func(*args, **kwargs)
-                    await tx.commit()
+                    await db_or_tx.on_transaction_commit.run(tx)
+                    await tx._commit()
                 except FoundError as exc:
                     await tx._on_error(exc.code)
+                    await db_or_tx.on_transaction_rollback.run(tx)
                 else:
                     return out
 
@@ -501,8 +511,15 @@ def transactional(func):
 
 
 class Database(BaseFound):
-    def __init__(self, pointer):
-        self._pointer = pointer
+
+    def __init__(self, cluster_file=None):
+        pointer = ffi.new("FDBDatabase **")
+        lib.fdb_create_database(cluster_file or ffi.NULL, pointer)
+        self._pointer = pointer[0]
+        self.on_transaction_begin = Hook()
+        self.on_transaction_commit = Hook()
+        self.on_transaction_rollback = Hook()
+        self.state = defaultdict(dict)
 
     def __del__(self):
         lib.fdb_database_destroy(self._pointer)
@@ -535,66 +552,6 @@ class Database(BaseFound):
         return out
 
 
-@ffi.callback("void(FDBFuture *, void *)")
-def on_create_database(fdb_future, aio_future):
-    aio_future = ffi.from_handle(aio_future)
-    pointer = ffi.new("FDBDatabase **")
-    error = lib.fdb_future_get_database(fdb_future, pointer)
-    if error == 0:
-        _loop.call_soon_threadsafe(aio_future.set_result, pointer[0])
-    else:
-        _loop.call_soon_threadsafe(aio_future.set_exception, FoundError(error))
-    lib.fdb_future_destroy(fdb_future)
-
-
-class Cluster(BaseFound):
-    def __init__(self, pointer):
-        self._pointer = pointer
-
-    @property
-    def options():
-        raise NotImplementedError()  # TODO
-
-    def __del__(self):
-        lib.fdb_cluster_destroy(self._pointer)
-
-    async def open(self, name):
-        name = name.encode("utf-8")
-        fdb_future = lib.fdb_cluster_create_database(self._pointer, name, len(name))
-        aio_future = _loop.create_future()
-        handle = ffi.new_handle(aio_future)
-        lib.fdb_future_set_callback(fdb_future, on_create_database, handle)
-        pointer = await aio_future
-        return Database(pointer)
-
-
-@ffi.callback("void(FDBFuture *, void *)")
-def on_create_cluster(fdb_future, aio_future):
-    aio_future = ffi.from_handle(aio_future)
-    pointer = ffi.new("FDBCluster **")
-    error = lib.fdb_future_get_cluster(fdb_future, pointer)
-    if error == 0:
-        _loop.call_soon_threadsafe(aio_future.set_result, pointer[0])
-    else:
-        _loop.call_soon_threadsafe(aio_future.set_exception, FoundError(error))
-    lib.fdb_future_destroy(fdb_future)
-
-
-async def create_cluster(cluster_file=None):
-    fdb_future = lib.fdb_create_cluster(cluster_file or ffi.NULL)
-    aio_future = _loop.create_future()
-    handle = ffi.new_handle(aio_future)
-    lib.fdb_future_set_callback(fdb_future, on_create_cluster, handle)
-    pointer = await aio_future
-    return Cluster(pointer)
-
-
-_clusters = dict()
-_databases = dict()
-
-cache_lock = threading.Lock()
-
-
 async def open(cluster_file=None):
     """Open the database specified by `cluster_file` or the default
     cluster indicated by the fdb.cluster file in a platform-specific
@@ -604,22 +561,6 @@ async def open(cluster_file=None):
         if _network_thread is None:
             init()
 
-    with cache_lock:
-        try:
-            cluster = _clusters[cluster_file]
-        except KeyError:
-            cluster = _clusters[cluster_file] = await create_cluster(cluster_file)
+    db = Database(cluster_file)
 
-        # in the 510 bindings, it is apparently possible to open different
-        # databases in the same cluster but that is not supported by
-        # the underlying client API. YAGNI for the time being. We hardcode
-        # the database name and do not expose it in the public API
-        database_name = "DB"
-        try:
-            db = _databases[(cluster_file, database_name)]
-        except KeyError:
-            db = _databases[(cluster_file, database_name)] = await cluster.open(
-                database_name
-            )
-
-        return db
+    return db
